/* $Id: cholesky.h,v 1.3 2006/01/18 11:54:49 cvsroot Exp $ */

/* GenLib2 - General Library 2
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef GENLIB2_CHOLESKY_INCLUDED
#define GENLIB2_CHOLESKY_INCLUDED

#include "vector.h"
#include "matrix.h"

namespace GenLib2
{

/** Cholesky decomposition.
 *  Cholesky decomposition: C'C = A. Only for symmetric and
 *  positive definite matrices. Only upper part of A is accessed;
 *  this part is not modified. The Cholesky factor C is returned
 *  in lower part of A while the diagonal elements are stored
 *  in the decomposition vector d. If the decomposition fails
 *  the matrix is not positive definite. In fact,
 *  Cholesky decomposition is a way to test wheather a symmetic
 *  matrix is positive definite.
 *  For more information see Press et al, Numerical Recipes in C.
 *
 *  @param A (IN/OUT) array with matrix A in the upper part and
 *  matrix C in the lower part
 *  @param d (OUT) decomposition vector
 *  @return info: info = 0, OK; info = -i, invalid i-th argument;
 *  info = i, decomposition failed in row i-1
 */
template<class T>
int chol_dec(Matrix<T>& A, Vector<T>& d)
{
	int m = A.rows();
	int n = A.cols();
	if (m != n) return -1; // not square!!!

	register int i, j, k;
	register T s;

	d.newsize(m);
	for (i = 0; i < m; i++)
		{
		s = 0.;
		for (k = 0; k < i; k++)
			s += A(i,k) * A(i,k);

		if ((d(i) = A(i,i) - s) <= 0.)
			return (i + 1);

		d(i) = sqrt(d(i));

		for (j = i + 1; j < m; j++)
			{
			s = A(i,j);
			for (k = 0; k < i; k++)
				s -= A(i,k) * A(j,k);

			A(j,i) = s / d(i);
			}
		}

	return 0;
}


/** Solving equation system.
 *  Solving after decomposition by backsubstitution: C'g = b and Cx = g.
 *  Vector d for saving the diagonal elements of C.
 *  Apply previously the decomposition function chol_dec.
 *
 *  @param A (IN) array A with matrix C in the lower part 
 *  @param d (IN) decomposition vector
 *  @param b (IN) right hand side vector
 *  @param x (OUT) solution vector
 *  @return info; info = 0, OK; info = -i, invalid i-th argument.
 */
template<class T>
int chol_sol(const Matrix<T>& A, const Vector<T>& d,
const Vector<T>& b, Vector<T>& x)
{
	int m = A.rows();
	int n = A.cols();
	if (m != n) return -1; // not square!!!
	if (d.size() != m) return -2; // different length!!!
	if (b.size() != m) return -3; // different length!!!

	register int i, k;
	register T s;

	x.newsize(m);
	for (i = 0; i < m; i++)
		{
		s = b(i);
		for (k = i - 1; k >= 0; k--)
			s -= A(i,k) * x(k);

		x(i) = s / d(i);
		}

	for (i = m - 1; i > -1; i--)
		{
		s = x(i);
		for (k = m - 1; k > i; k--)
			s -= A(k,i) * x(k);

		x(i) = s / d(i);
		}
	
	return 0;
}


/** inversion 
 *  Inversion after decomposition: CQ = INV(C').
 *  Vector d for storing the diagonal elements of C.
 *  Apply previously the decomposition function chol_dec.
 *
 *  @param A (IN/OUT) array with matrix C in the lower part and
 *  inverse matrix Q in the upper part
 *  @param d (IN) decomposition vector
 *  @return info; info = -i, invalid i-th argument; info = 0, OK.
 */
template<class T>
int chol_inv(Matrix<T>& A, const Vector<T>& d)
{
	int m = A.rows();
	int n = A.cols();
	if (m != n) return -1; // not square!!!
	if (d.size() != m) return -2; // different length!!!

	register int i, j, k;
	register T s;

	for (i = m - 1; i > -1; i--)
		{
		s = 1. / d(i);

		for (k = i + 1; k < m; k++)
			s -= A(k,i) * A(i,k);

		A(i,i) = s / d(i);

		for (j = i - 1; j > -1; j--)
			{
			s = 0.;
			for (k = j + 1; k < i; k++)
				s -= A(k,j) * A(k,i);

			for (k = i; k < m; k++)
				s -= A(k,j) * A(i,k);
			
			A(j,i) = s / d(j);
			}
		}
	
	return 0;
}

} // namespace GenLib2

#endif /* GENLIB2_CHOLESKY_INCLUDED */
